from PIL import Image
from astropy.io import fits


def make_tif(data=[], metadata=[], file_name=''):
    '''create tif file'''
    new_image = Image.fromarray(data)
    new_image.save(file_name, tiffinfo=metadata)

def make_fits(data=[], file_name=''):
    '''create fits file'''
    fits.writeto(file_name, data, clobber=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan  6 15:42:58 2017

@author: harti and valsecchi
"""
from functions import read_data,cropped,createIm,normalization,saveIm,binning,oscillation,createIm_fft


path_ob = 'data/data_OB'
path_im = 'data/data_smp'
path_dc = 'data/DCs'

bin_fac = None                 # no binning either 1 or None, 2x2 binning: bin_fac = 2
norm_param = [3,5,20,40]
crop_param = [10,15,80,60]
oscillationParam = [30,1,1,1]
numberPeriods = 1
savingFolder = 'folder'

im,ob = read_data(path_im,path_ob,path_dc)
#im,ob=normalization(im,ob,*norm_param)
#oscillation(im,ob,folder=savingFolder,*oscillationParam)
#im,ob = cropped(im,ob,*crop_param)
#im, ob = binning(im,ob,bin_fac)
ti, dpci, dfi, vis_map = createIm(im,ob,numberPeriods)
saveIm(ti, dpci, dfi, vis_map,name='period2',folder=savingFolder,overWrite=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan  6 15:10:52 2017

@author: harti and valsecchi
"""

import os
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib import gridspec
from os import makedirs
from astropy.io import fits 
import h5py
from pathlib import Path
from scipy.signal import medfilt,wiener



def readRead(path,dc=0):
    """
    Function to read data from the specified path, it can read FITS, TIFF and HDF.
    
    Parameters
    ----------
    path : string_like
            Path of the input file with his extention.
    dc : array_like
            An array containing the dark current data.
    
    Returns
    -------
    im_a1 : array_like
            An array containing the read data.
    
    Notes
    -----
    In case of corrupted header it skips the header and reads the raw data.
    For the HDF format you need to specify the hierarchy.
    """
    my_file = Path(path)
    if my_file.is_file():
        im_a1 = []
        if path.lower().endswith('.fits'):
            try:
                temp = fits.open(path,ignore_missing_end=True)[0].data
                if len(temp.shape) == 3:
                    temp = temp.reshape(temp.shape[1:])                
                im_a1.append(temp)  
            except OSError:
                import fitsio
                im_a1.append(fitsio.read(path))
        elif path.lower().endswith(('.tiff','.tif')) :
            im_a1.append(np.asarray(Image.open(path)))
        elif path.lower().endswith(('.hdf','.h4','.hdf4','.he2','h5','.hdf5','.he5')): 
            # change here the hierarchy 
            hdf = h5py.File(path,'r')['entry']['data']['data'].value    
            for iScan in hdf:
                im_a1.append(iScan)
        else:
            raise OSError('file extension not yet implemented....Do it your own way!')     
        im_a1 = np.asarray(im_a1)-dc
        return im_a1
    else:
        raise OSError('the path does not exist')
    
def read_data(path_im,path_ob,path_dc):
    """
    Function to read data from the specified folders, it can read FITS, TIFF and HDF.

    Parameters
    ----------
    path_im : string_like
            Path of the projection folder.
            
    path_ob : string_like
            Path of the open beam folder.
            
    path_dc : string_like
            Path of the dark current folder.
            
    Returns
    -------
    stack_im_ar : array_like
            A stack containing all the projections.
    
    stack_ob : array_like
            A stack containing all the references.
    
    Notes
    -----   
    File format supported: ['.fits','.tiff','.tif','.hdf','.h4','.hdf4','.he2','h5','.hdf5','.he5']
    """
#    Dark current
    imExt = ['.fits','.tiff','.tif','.hdf','.h4','.hdf4','.he2','h5','.hdf5','.he5']
    if path_dc:
        im_a1 = []
        filenames_dc = [name for name in os.listdir(path_dc) if name.lower().endswith(tuple(imExt))]
        filenames_dc.sort()
        for name in filenames_dc:
            full_path_name = path_dc+'/'+name
#            print(full_path_name)
            im_a1.append(readRead(full_path_name))
        im_a1 = np.asarray(im_a1)
        im_a1 = np.sum(im_a1,axis=0)/np.shape(im_a1)[0]
    
#    Open beam
    filenames_ob = [name for name in os.listdir(path_ob) if name.lower().endswith(tuple(imExt))]
    filenames_ob.sort()
    stack_ob = []
    for name in filenames_ob:
        full_path_name = path_ob+'/'+name
        if path_dc:
            stack_ob.append(readRead(full_path_name,im_a1)) #with dc
        else:
            stack_ob.append(readRead(full_path_name))   #without dc
    stack_ob = np.concatenate(stack_ob)

#    Projections
    filenames_im = [name for name in os.listdir(path_im) if name.lower().endswith(tuple(imExt))]
    filenames_im.sort()
    stack_im_ar = []
    for name in filenames_im:
        full_path_name = path_im+'/'+name
        if path_dc:
            stack_im_ar.append(readRead(full_path_name,im_a1)) #with dc
        else:
            stack_im_ar.append(readRead(full_path_name))   #without dc
    stack_im_ar = np.concatenate(stack_im_ar)
    if np.shape(stack_im_ar) != np.shape(stack_ob):
            raise ValueError('Data and open beam have different shapes')
        
    return stack_im_ar,stack_ob


xROI,yROI,widthROI,heightROI=10,10,35,35 #parameter for roi
   
def roi(im,xROI,yROI,widthROI,heightROI,show=False,titleOne='Original image with selected ROI',titleTwo='ROI',shape=False):
    """
    roi() takes a single image and crops it, (xROI,yROI) is the upper left-hand corner of the rectangle in pixel units.
    
    Parameters
    ----------
    im : array_like
        Input image.
        
    xROI : int_like
        x position of the upper left-hand corner in pixel units.
    
    yROI : int_like
        y position of the upper left-hand corner in pixel units.
    
    widthROI : int_like
        width of the recangle in pixel units.
    
    heightROI : int_like
        height of the recangle in pixel units.
    
    show : boolean
        If True it shows the ROI and its position.
        
    titleOne : string_like
        Title of the ROI position.
        
    titleTwo : string_like
        Title of the ROI.
        
    shape : boolean
        If True it shows the size of the image.
        
    Returns
    -------
    imROI : array_like
        Return the ROI.
    
    Notes
    -----
    """
    if shape:
        print(im.shape)
    if (0<=xROI<=im.shape[1] and 0<=xROI+widthROI<=im.shape[1] and 0<=yROI<=im.shape[0] and 0<=yROI+heightROI<=im.shape[0]):
        imROI = im[yROI:yROI+heightROI,xROI:xROI+widthROI]
        if show:
            vmin,vmax=np.mean(im)-2*np.std(im),np.mean(im)+2*np.std(im)
            print(vmax, vmin)
            cmap='gray'
            fig = plt.figure(figsize=(15,10)) 
            gs = gridspec.GridSpec(1, 2,width_ratios=[4,1],height_ratios=[1,1]) 
            ax = plt.subplot(gs[0])
            ax2 = plt.subplot(gs[1])
            ax.imshow(im,vmin=vmin, vmax=vmax,interpolation='nearest',cmap=cmap)
            rectNorm = patches.Rectangle((xROI,yROI),widthROI,heightROI,linewidth=1,edgecolor='m',facecolor='none')
            ax.add_patch(rectNorm)
            ax.set_title(titleOne)
            ax2.imshow(im,vmin=vmin, vmax=vmax,interpolation='nearest',cmap=cmap)
            ax2.set_title(titleTwo)
            ax2.set_xlim([xROI,xROI+widthROI])
            ax2.set_ylim([yROI+heightROI,yROI])
            plt.tight_layout()
            plt.show()
            plt.close('all')
      
        return(imROI)
    else:
        print('!!!WARNING!!! \nROI out of range')


def cropped(stack_im,stack_ob,xROI=xROI,yROI=yROI,widthROI=widthROI,heightROI=heightROI,show=True):
    """
    Crop the stacks to the selected ROI 
   
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
        
    xROI : int_like
        x position of the upper left-hand corner in pixel units.
    
    yROI : int_like
        y position of the upper left-hand corner in pixel units.
    
    widthROI : int_like
        width of the recangle in pixel units.
    
    heightROI : int_like
        height of the recangle in pixel units.
    
    show : boolean
        If True it shows the ROI and its position.
    
    Returns
    -------
    stack_im_ar : array_like
            A stack containing all the cropped projections.
    
    stack_ob_ar : array_like
            A stack containing all the cropped references.  
            
    Notes
    -----
    
    """
    stack_im_ar = [roi(im=stack_im[0],xROI=xROI,yROI=yROI,widthROI=widthROI,heightROI=heightROI,show=show,titleTwo='Cropped region',shape=True)]
    for i in stack_im[1:]:
        stack_im_ar.append(roi(im=i,xROI=xROI,yROI=yROI,widthROI=widthROI,heightROI=heightROI,show=False))    
    stack_ob_ar = [roi(im=i,xROI=xROI,yROI=yROI,widthROI=widthROI,heightROI=heightROI,show=False) for i in stack_ob]
    
    return(np.asarray(stack_im_ar),np.asarray(stack_ob_ar))

    
def normalization(stack_im,stack_ob,xROI=xROI,yROI=yROI,widthROI=widthROI,heightROI=heightROI,show=True):
    """
    Normalize the stacks to the selected ROI 
   
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
        
    xROI : int_like
        x position of the upper left-hand corner in pixel units.
    
    yROI : int_like 
        y position of the upper left-hand corner in pixel units.
    
    widthROI : int_like
        width of the recangle in pixel units.
    
    heightROI : int_like
        height of the recangle in pixel units.
    
    show : boolean
        If True it shows the ROI and its position.
    
    Returns
    -------
    stack_im_ar : array_like
            A stack containing all the normalized projections.
    
    stack_ob_ar : array_like
            A stack containing all the normalized references.  
            
    Notes
    -----
    
    """
    area = abs(widthROI*heightROI)  
    stack_im_ar = []    
    roi(stack_im[0],xROI,yROI,widthROI,heightROI,show,titleTwo='Area for normalization')
    stack_im_ar = [l/(l[yROI:yROI+heightROI+1,xROI:xROI+widthROI+1].sum()/area) for l in stack_im]   
    stack_ob_ar = [l/(l[yROI:yROI+heightROI+1,xROI:xROI+widthROI+1].sum()/area) for l in stack_ob] 
    
    return(np.asarray(stack_im_ar),np.asarray(stack_ob_ar))

def oscillation(stack_im,stack_ob,xROI=xROI,yROI=yROI,widthROI=widthROI,heightROI=heightROI,repeatedPeriod=False,folder=False,show=True):
    """
    Oscillation plot analysis of the selected ROI.
   
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
        
    xROI : int_like
        x position of the upper left-hand corner in pixel units.
    
    yROI : int_like
        y position of the upper left-hand corner in pixel units.
    
    widthROI : int_like
        width of the recangle in pixel units.
    
    heightROI : int_like
        height of the recangle in pixel units.
    
    repeatedPeriod : boolean
        If True it plots the oscillation twice.
        
    folder : string_like or False
        Path of the folder where to save the oscillation plot, if False it doesn't save the plot.
    
    show : boolean
        If True it shows the ROI and its position.
    
    Returns
    -------
    Oscillation plot.  
            
    Notes
    -----  
    """
    titleOne='Area for oscillation'
    titleTwo='Oscillation plot'
    stack_ob_ar = [l[yROI:yROI+heightROI,xROI:xROI+widthROI].mean() for l in stack_ob] 
    stack_im_ar = [l[yROI:yROI+heightROI,xROI:xROI+widthROI].mean() for l in stack_im] 

    if repeatedPeriod:
        stack_ob_ar += stack_ob_ar
        stack_im_ar += stack_im_ar
        titleTwo += ' repetead period'
#    PLOT oscillation
    im = stack_im[0]
    if (0<=xROI<=im.shape[1] and 0<=xROI+widthROI<=im.shape[1] and 
        0<=yROI<=im.shape[0] and 0<=yROI+heightROI<=im.shape[0]):
        vmin,vmax=im.min(),im.max()
        cmap='gray'
        fig = plt.figure(figsize=(15,10)) 
        gs = gridspec.GridSpec(1,2,width_ratios=[2,1],height_ratios=[1,1]) 
        ax = plt.subplot(gs[0])
        ax2 = plt.subplot(gs[1])
        ax.imshow(im,vmin=vmin, vmax=vmax,interpolation='nearest',cmap=cmap)
        rectNorm = patches.Rectangle((xROI,yROI),widthROI,heightROI,linewidth=1,edgecolor='m',facecolor='none')
        ax.add_patch(rectNorm)
        ax.set_title(titleOne)
        rangeim = range(1,len(stack_im_ar)+1)
        ax2.plot(rangeim,stack_im_ar,color='g',label='data')
        ax2.scatter(rangeim,stack_im_ar,marker='*',color='g')
        rangeob = range(1,len(stack_ob_ar)+1)
        ax2.plot(rangeob,stack_ob_ar,color='b',label='ob')
        ax2.scatter(rangeob,stack_ob_ar,color='b')
        ax2.legend(loc=1, shadow=True)
        ax2.set_title(titleTwo)
        ax2.set_xlim((0,len(stack_ob_ar)+2))
        ax2.grid(True)
        plt.tight_layout()
        if show:
            plt.show()
        if folder:
            if not os.path.exists(folder):
                makedirs(folder) 
                print('files saved in folder: ',str(folder))
            fig.savefig(str(folder)+'/oscillationPlot_X'+str(xROI)+'Y'+str(yROI)+'.png', bbox_inches='tight')
        plt.close('all')
    else:
        print('!!!WARNING!!! \nROI out of range')
    

def matrix(stack_im,numberPeriods):
    """
    Take the stack of OBs and projections to create TI, DPCI, DFI and visiblity map.
    The algorithm used in this step is based on Marathe et al.
    (2014) http://dx.doi.org/10.1063/1.4861199.
    
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
        
    numberPeriods : float_like
        Numbers or fraction of stepped period.
    
    Returns
    -------
    a0 : array_like
        Offset.
        
    a1 : array_like
        Amplitude.
    
    phi : array_like
        Phase.
    
    Notes
    -----
    """
    shapeStack = np.shape(stack_im)
    B = np.zeros((shapeStack[0],3))  
    
    stack_imReshaped = np.reshape(stack_im,[shapeStack[0],shapeStack[1]*shapeStack[2]])
    rangeStack = range(shapeStack[0])
    for j in rangeStack:
        B[j][0] = 1.0
        B[j][1] = np.cos(2*np.pi*rangeStack[j]*numberPeriods/(shapeStack[0]-1))
        B[j][2] = np.sin(2*np.pi*rangeStack[j]*numberPeriods/(shapeStack[0]-1))
    B = np.matrix(B)
    G = (B.T * B).I * B.T
    A = (G*stack_imReshaped)
    offSet,absoluteAmpl,absPhase = A[0,:],A[1,:],A[2,:]
    a0 = np.reshape(np.asarray(offSet),[shapeStack[1],shapeStack[2]])
    a1 = np.reshape(np.sqrt(np.asarray(absoluteAmpl)**2+np.asarray(absPhase)**2),[shapeStack[1],shapeStack[2]])
    phi = np.reshape(np.arctan((np.asarray(absPhase)/np.asarray(absoluteAmpl))),[shapeStack[1],shapeStack[2]])
    
    return a0,a1,phi
     
def reductionMatrix(stack_im,stack_ob,numberPeriods):
    """
    Apply matrix() to the projections and open beam stacks.
   
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
        
    numberPeriods : float_like
        Numbers or fraction of stepped period.
    
    Returns
    -------
    Return an array_like containing the a0, a1, phi of the projections and open beam stack.
    
    Notes
    -----
    """
    return (matrix(stack_im,numberPeriods),matrix(stack_ob,numberPeriods))

def createIm_fft(stack_im,stack_ob):
    """
    Apply the fourier component analysis to retrieve dfi, ti, dpci and visibility map.

    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
    
    Returns
    -------
    ti : array_like
        Reconstructed fft trasmission array.
        
    dpci : array_like
        Reconstructed fft DPCI array.
        
    dfi : array_like
        Reconstructed fft DFI array.
        
    visi : array_like
        Reconstructed fft visibility array.
    Notes
    -----
    """
    ## Projection
    shapeStack_im = np.shape(stack_im)
    stack_imReshaped = np.reshape(stack_im,[shapeStack_im[0],shapeStack_im[1]*shapeStack_im[2]])
    steps = len(stack_imReshaped[:,:])
    outfft_a_im = []
    outfft_b_im = []
    outfft_c_im = []
    for i in range(len(stack_imReshaped[0,:])):
        fft_osc = np.fft.rfft(stack_imReshaped[:,i])/(steps-1)
        outfft_a_im.append(fft_osc[0])
        outfft_b_im.append(2*np.sqrt(np.real(fft_osc[1])**2+np.imag(fft_osc[1])**2))
        outfft_c_im.append(-np.angle(fft_osc[1]))
    outfft_a_im = np.reshape(outfft_a_im, [shapeStack_im[1],shapeStack_im[2]])
    outfft_b_im = np.reshape(outfft_b_im, [shapeStack_im[1],shapeStack_im[2]])
    outfft_c_im = np.reshape(outfft_c_im, [shapeStack_im[1],shapeStack_im[2]])

    ## open beam
    shapeStack_ob = np.shape(stack_ob)
    stack_obReshaped = np.reshape(stack_ob,[shapeStack_ob[0],shapeStack_ob[1]*shapeStack_ob[2]])
    steps = len(stack_obReshaped[:,:])
    outfft_a_ob = []
    outfft_b_ob = []
    outfft_c_ob = []
    for i in range(len(stack_obReshaped[0,:])):
        fft_osc = np.fft.rfft(stack_obReshaped[:,i])/(steps-1)
        outfft_a_ob.append(fft_osc[0])
        outfft_b_ob.append(2*np.sqrt(np.real(fft_osc[1])**2+np.imag(fft_osc[1])**2))
        outfft_c_ob.append(-np.angle(fft_osc[1]))
    outfft_a_ob = np.reshape(outfft_a_ob, [shapeStack_ob[1],shapeStack_ob[2]])
    outfft_b_ob = np.reshape(outfft_b_ob, [shapeStack_ob[1],shapeStack_ob[2]])
    outfft_c_ob = np.reshape(outfft_c_ob, [shapeStack_ob[1],shapeStack_ob[2]])
    ti = np.real(outfft_a_im)/np.real(outfft_a_ob)
    dpci = outfft_c_im - outfft_c_ob
    dfi = (np.real(outfft_b_im)*np.real(outfft_a_ob))/(np.real(outfft_b_ob)*np.real(outfft_a_im))
    visi = np.real(outfft_b_ob)/np.real(outfft_a_ob)
    
    return ti, dpci, dfi, visi


def createIm(stack_im,stack_ob,numberPeriods):
    """
    Process the projections and open beam stacks to retrieve dfi, ti, dpci and visibility map.
    
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
        
    numberPeriods : float_like
        Numbers or fraction of stepped period.
        
    Returns
    -------
    TI : array_like
        Reconstructed trasmission array.
        
    DPCI : array_like
        Reconstructed DPCI array.
        
    DFI : array_like
        Reconstructed DFI array.
        
    VIS_map : array_like
        Reconstructed visibility array.
    
    Notes
    -----
    """
    imParam,obParam = reductionMatrix(stack_im,stack_ob,numberPeriods)
    TI = np.divide(imParam[0],obParam[0])
    DPCI = imParam[2]-obParam[2]
    DPCI = np.arctan(np.tan(DPCI))
    DFI = np.divide(np.divide(imParam[1],imParam[0]),np.divide(obParam[1],obParam[0]))
    VIS_map = np.divide(obParam[1],obParam[0])
    
    return TI, DPCI, DFI, VIS_map
    
def saveIm(ti,dpci,dfi,vis_map,name='name',folder='folder',overWrite=False):
    """
    Save the images to specific folder.
    
    Parameters
    ----------
    ti : array_like
        Input reconstructed trasmission array.
        
    dpci : array_like
        Input reconstructed DPCI array.
        
    dfi : array_like
        Input reconstructed DFI array.
        
    vis_map : array_like
        Input reconstructed visibility array.
    
    name : sring_like
        Specific name for the images.
        
    folder : string_like
        Path of the folder where to save the reconstructed images.
        
    overWrite : boolean
        If it's True it overwrite the images with the same name.
        
    Returns
    -------
    
    Notes
    -----
    """
    if not os.path.exists(folder):
        makedirs(folder) 
    print('files saved in folder: ',folder)
    fits.writeto(folder+'/ti_'+str(name)+'.fits',ti,clobber=overWrite)
    fits.writeto(folder+'/dpci_'+str(name)+'.fits',dpci,clobber=overWrite)
    fits.writeto(folder+'/dfi_'+str(name)+'.fits',dfi,clobber=overWrite)
    fits.writeto(folder+'/visi_'+str(name)+'.fits',vis_map,clobber=overWrite)
    
def binning(stack_im,stack_ob,bin_fac=None):
    """
    Bin the projections and open beam stacks. 
    
    Parameters
    ----------
    stack_im : array_like
        Input stack of projections.
    
    stack_ob : array_like
        Input stack of open beam.
        
    bin_fac : int_like
        Cluster of pixels into a single pixel.
    
    Returns
    -------
    stack_im_bin : array_like
            A stack containing all the binned projections.
    
    stack_ob_bin : array_like
            A stack containing all the binned references.  
    
    Notes
    -----
    """
    num_im,x_im,y_im = np.shape(stack_im)
    stack_im_tmp = list()
    if bin_fac:
        for i in range(num_im):
            t = stack_im[i]
            t_resh = t.reshape(int(x_im/bin_fac), bin_fac, int(y_im/bin_fac), bin_fac)
            stack_im_tmp.append(t_resh.mean(axis=3).mean(axis=1))
    
        stack_im_bin = np.asarray(stack_im_tmp) 
    else:
        stack_im_bin = stack_im
        
    num_ob,x_ob,y_ob = np.shape(stack_ob)
    stack_ob_tmp = list()
    if bin_fac:
        for i in range(num_ob):
            t = stack_ob[i]
            t_resh = t.reshape(int(x_ob/bin_fac), bin_fac, int(y_ob/bin_fac), bin_fac)
            stack_ob_tmp.append(t_resh.mean(axis=3).mean(axis=1))
    
        stack_ob_bin = np.asarray(stack_ob_tmp) 
    else:
        stack_ob_bin = stack_ob
        
    return stack_im_bin, stack_ob_bin



from pathlib import Path
import numpy as np
import os
import copy
import time
from scipy.ndimage import convolve

from NeuNorm.loader import load_hdf, load_tiff, load_fits
from NeuNorm.exporter import make_fits, make_tif
from NeuNorm.roi import ROI
from NeuNorm._utilities import get_sorted_list_images, average_df


class Normalization(object):

    gamma_filter_threshold = 0.1
    
    def __init__(self):
        self.shape = {'width': np.NaN,
                      'height': np.NaN}
        self.dict_image = { 'data': [],
                            'oscilation': [],
                            'file_name': [],
                            'metadata': [],
                            'shape': self.shape.copy()}
        self.dict_ob = {'data': [],
                        'oscilation': [],
                        'metadata': [],
                        'file_name': [],
                        'data_mean': [],
                        'shape': self.shape.copy()}
        self.dict_df = {'data': [],
                        'metadata': [],
                        'data_average': [],
                        'file_name': [],
                        'shape': self.shape.copy()}

        __roi_dict = {'x0': np.NaN,
                      'x1': np.NaN,
                      'y0': np.NaN,
                      'y1': np.NaN}
        self.roi = {'normalization': __roi_dict.copy(),
                    'crop': __roi_dict.copy()}

        self.__exec_process_status = {'df_correction': False,
                                      'normalization': False,
                                      'crop': False,
                                      'oscillation': False,
                                      'bin': False}

        self.data = {}
        self.data['sample'] = self.dict_image
        self.data['ob'] = self.dict_ob
        self.data['df'] = self.dict_df
        self.data['normalized'] = []
        self.export_file_name = []
    
    def load(self, file='', folder='', data=[], data_type='sample', 
             gamma_filter=True, notebook=False):
        '''
        Function to read individual files or entire files from folder specify for the given
        data type
        
        Parameters:
           file: full path to file
           folder: full path to folder containing files to load
           data: numpy array
           data_type: ['sample', 'ob', 'df]
           gamma_filter: boolean (default True) apply or not gamma filtering to the data loaded
           notebooks: boolean (default False) turn on this option if you run the library from a
             notebook to have a progress bar displayed showing you the progress of the loading
           
        Algorithm won't be allowed to run if any of the main algorithm have been run already, such as
        oscillation, crop, binning, df_correction.

        '''
        list_exec_flag = [_flag for _flag in self.__exec_process_status.values()]
        box1 = None
        if True in list_exec_flag:
            raise IOError("Operation not allowed as you already worked on this data set!")
       
        if notebook:
            from ipywidgets import widgets
            from IPython.core.display import display
        
        if not file == '':
            if isinstance(file, str):
                self.load_file(file=file, data_type=data_type)
            elif isinstance(file, list):
                if notebook:
                    # turn on progress bar
                    _message = "Loading {}".format(data_type)
                    box1 = widgets.HBox([widgets.Label(_message,
                                                       layout=widgets.Layout(width='20%')),
                                         widgets.IntProgress(max=len(file)),
                                         widgets.Label("Time remaining:",
                                                       layout=widgets.Layout(width='10%')),
                                         widgets.Label(" >> calculating << ")])
                    display(box1)
                    w1 = box1.children[1]                    
                    time_remaining_ui = box1.children[-1]

                start_time = time.time()
                for _index, _file in enumerate(file):
                    self.load_file(file=_file, data_type=data_type)
                    if notebook:
                        w1.value = _index+1
                        end_time = time.time()
                        takes_its_going_to_take = self.calculate_how_long_its_going_to_take(index_we_are=_index + 1,
                                                                                            time_it_took_so_far=end_time - start_time,
                                                                                            total_number_of_loop=len(file))
                        time_remaining_ui.value = "{}".format(takes_its_going_to_take)

                if notebook:
                    box1.close()

        if not folder == '':
            # load all files from folder
            list_images = get_sorted_list_images(folder=folder)
            if notebook:
                # turn on progress bar
                _message = "Loading {}".format(data_type)
                box1 = widgets.HBox([widgets.Label(_message,
                                                   layout=widgets.Layout(width='20%')),
                                     widgets.IntProgress(max=len(list_images)),
                                     widgets.Label("Time remaining:",
                                                   layout=widgets.Layout(width='10%')),
                                     widgets.Label(" >> calculating << ")])
                display(box1)
                w1 = box1.children[1]
                time_remaining_ui = box1.children[-1]

            start_time = time.time()
            for _index, _image in enumerate(list_images):
                full_path_image = os.path.join(folder, _image)
                self.load_file(file=full_path_image, data_type=data_type, gamma_filter=gamma_filter)
                if notebook:
                    # update progress bar
                    w1.value = _index+1
                    end_time = time.time()
                    takes_its_going_to_take = self.calculate_how_long_its_going_to_take(index_we_are=_index+1,
                                                                                        time_it_took_so_far=end_time-start_time,
                                                                                        total_number_of_loop=len(list_images))
                    time_remaining_ui.value = "{}".format(takes_its_going_to_take)

            if notebook:
                box1.close()
        
        if not data == []:
            self.load_data(data=data, data_type=data_type)

    def calculate_how_long_its_going_to_take(self, index_we_are=-1, time_it_took_so_far=0, total_number_of_loop=1):
        time_per_loop = time_it_took_so_far / index_we_are
        total_time_it_will_take = time_per_loop * total_number_of_loop
        time_left = total_time_it_will_take - time_per_loop * index_we_are

        # convert to nice format h mn and seconds
        m, s = divmod(time_left, 60)
        h, m = divmod(m, 60)

        if h == 0:
            if m == 0:
                return "%02ds" %(s)
            else:
                return "%02dmn %02ds" %(m, s)
        else:
            return "%dh %02dmn %02ds" % (h, m, s)

    def load_data(self, data=[], data_type='sample', notebook=False):
        '''Function to save the data already loaded as arrays

        Paramters:
        ==========
        data: np array 2D or 3D 
        data_type: string ('sample')
        notebook: boolean (default False) turn on this option if you run the library from a
             notebook to have a progress bar displayed showing you the progress of the loading
        '''
        if notebook:
            from ipywidgets import widgets
            from IPython.core.display import display

        if len(np.shape(data)) > 2:
            if notebook:
                _message = "Loading {}".format(data_type)
                box1 = widgets.HBox([widgets.Label(_message,
                                                               layout=widgets.Layout(width='20%')),
                                                 widgets.IntProgress(max=len(list_images))])
                display(box1)
                w1 = box1.children[1]   

            for _index, _data in enumerate(data):
                _data = _data.astype(float)
                self.__load_individual_data(data=_data, data_type=data_type)
                if notebook:
                    # update progress bar
                    w1.value = _index+1

            if notebook:
                box1.close()
                    
        else:
            data = data.astype(float)
            self.__load_individual_data(data=data, data_type=data_type)
            
    def __load_individual_data(self, data=[], data_type='sample'):
        self.data[data_type]['data'].append(data)
        index = len(self.data[data_type]['data'])
        self.data[data_type]['file_name'].append("image_{:04}".format(index))
        self.save_or_check_shape(data=data, data_type=data_type)        
        
    def load_file(self, file='', data_type='sample', gamma_filter=True):
        """
        Function to read data from the specified path, it can read FITS, TIFF and HDF.
    
        Parameters
        ----------
        file : string_like
            Path of the input file with his extention.
        data_type: ['sample', 'df']
        gamma_filter: Boolean (default True) apply or not gamma filtering
    
        Notes
        -----
        In case of corrupted header it skips the header and reads the raw data.
        For the HDF format you need to specify the hierarchy.
        """
    
        my_file = Path(file)
        if my_file.is_file():
            metadata = {}
            if file.lower().endswith('.fits'):
                data = np.array(load_fits(my_file), dtype=np.float)
            elif file.lower().endswith(('.tiff','.tif')) :
                [data, metadata] = load_tiff(my_file)
                data = np.array(data, dtype=np.float)
            elif file.lower().endswith(('.hdf','.h4','.hdf4','.he2','h5','.hdf5','.he5')):
                data = np.array(load_hdf(my_file), dtype=np.float)
            else:
                raise OSError('file extension not yet implemented....Do it your own way!')     

            if gamma_filter:
                data = self._gamma_filtering(data=data)

            data = np.squeeze(data)

            self.data[data_type]['data'].append(data)
            self.data[data_type]['metadata'].append(metadata)
            self.data[data_type]['file_name'].append(file)
            self.save_or_check_shape(data=data, data_type=data_type)

        else:
            raise OSError("The file name does not exist")

    def _gamma_filtering(self, data=[]):
        '''perform gamma filtering on the data
        
        Algorithm looks for all the very hight counts
        
        if self.gamma_filter_threshold * pixels[x,y] > mean_counts(data) then this pixel counts
        is replaced by the average value of the 8 pixels surrounding him
        
        Parameters:
        ===========
        data: numpy 2D array
        
        Returns:
        =======
        numpy 2D array 
        '''
        if data == []:
            raise ValueError("Data array is empty!")

        data_gamma_filtered = np.copy(data)
            
        # find mean counts
        mean_counts = np.mean(data_gamma_filtered)
        
        thresolded_data_gamma_filtered = data_gamma_filtered * self.gamma_filter_threshold
        
        # get pixels where value is above threshold
        position = []
        [height, width] = np.shape(data_gamma_filtered)
        for _x in np.arange(width):
            for _y in np.arange(height):
                if thresolded_data_gamma_filtered[_y, _x] > mean_counts:
                    position.append([_y, _x])
                    
        # convolve entire image using 3x3 kerne
        mean_kernel = np.array([[1,1,1], [1,0,1], [1,1,1]]) / 8.0
        convolved_data = convolve(data_gamma_filtered, mean_kernel, mode='constant')
        
        # replace only pixel above threshold by convolved data
        for _coordinates in position:
            [_y, _x] = _coordinates
            data_gamma_filtered[_y, _x] = convolved_data[_y, _x]
            
        return data_gamma_filtered        

    def save_or_check_shape(self, data=[], data_type='sample'):
        '''save the shape for the first data loaded (of each type) otherwise
        check the size match
    
        Raises:
        IOError if size do not match
        '''
        [height, width] = np.shape(data)
        if np.isnan(self.data[data_type]['shape']['height']):
            _shape = self.shape.copy()
            _shape['height'] = height
            _shape['width'] = width
            self.data[data_type]['shape'] = _shape
        else:
            _prev_width = self.data[data_type]['shape']['width']
            _prev_height = self.data[data_type]['shape']['height']
            
            if (not (_prev_width == width)) or (not (_prev_height == height)):
                raise IOError("Shape of {} do not match previous loaded data set!".format(data_type))

    def normalization(self, roi=None, force=False, force_mean_ob=False, notebook=False):
        '''normalization of the data 
                
        Parameters:
        ===========
        roi: ROI object or list of ROI objects that defines the region of the sample and OB that have to match
        in intensity
        force: boolean (default False) that if True will force the normalization to occur, even if it had been
        run before with the same data set
        notebook: boolean (default False) turn on this option if you run the library from a
             notebook to have a progress bar displayed showing you the progress of the loading

        Raises:
        =======
        IOError: if no sample loaded
        IOError: if no OB loaded
        IOError: if size of sample and OB do not match
        
        '''
        if not force:
            # does nothing if normalization has already been run
            if self.__exec_process_status['normalization']:
                return
        self.__exec_process_status['normalization'] = True
        
        # make sure we loaded some sample data
        if self.data['sample']['data'] == []:
            raise IOError("No normalization available as no data have been loaded")

        # make sure we loaded some ob data
        if self.data['ob']['data'] == []:
            raise IOError("No normalization available as no OB have been loaded")
              
        # make sure the data loaded have the same size
        if not self.data_loaded_have_matching_shape():
            raise ValueError("Data loaded do not have the same shape!")
              
        # make sure, if provided, roi has the right type and fits into the images
        b_list_roi = False
        if roi:
            if type(roi) is list:
                for _roi in roi:
                    if not type(_roi) == ROI:
                        raise ValueError("roi must be a ROI object!")
                    if not self.__roi_fit_into_sample(roi=_roi):
                        raise ValueError("roi does not fit into sample image!")
                b_list_roi = True

            elif not type(roi) == ROI:
                raise ValueError("roi must be a ROI object!")
            else:
                if not self.__roi_fit_into_sample(roi=roi):
                    raise ValueError("roi does not fit into sample image!")
        
        if notebook:
            from ipywidgets.widgets import interact
            from ipywidgets import widgets
            from IPython.core.display import display, HTML                

        # heat normalization algorithm
        _sample_corrected_normalized = []
        _ob_corrected_normalized = []

        if roi:

            if b_list_roi:

                _sample_corrected_normalized = []
                for _sample in self.data['sample']['data']:
                    sample_mean = []
                    for _roi in roi:
                        _x0 = _roi.x0
                        _y0 = _roi.y0
                        _x1 = _roi.x1
                        _y1 = _roi.y1
                        sample_mean.append(np.mean(_sample[_y0:_y1 + 1, _x0:_x1 + 1]))

                    full_sample_mean = np.mean(sample_mean)
                    _sample_corrected_normalized.append(_sample/full_sample_mean)

                _ob_corrected_normalized = []
                for _ob in self.data['ob']['data']:
                    ob_mean = []
                    for _roi in roi:
                        _x0 = _roi.x0
                        _y0 = _roi.y0
                        _x1 = _roi.x1
                        _y1 = _roi.y1
                        ob_mean.append(np.mean(_ob[_y0:_y1 + 1, _x0:_x1 + 1]))

                    full_ob_mean = np.mean(ob_mean)
                    _ob_corrected_normalized.append(_ob / full_sample_mean)

            else:
                _x0 = roi.x0
                _y0 = roi.y0
                _x1 = roi.x1
                _y1 = roi.y1

                _sample_corrected_normalized = [_sample / np.mean(_sample[_y0:_y1+1, _x0:_x1+1])
                                                for _sample in self.data['sample']['data']]
                _ob_corrected_normalized = [_ob / np.mean(_ob[_y0:_y1+1, _x0:_x1+1])
                                            for _ob in self.data['ob']['data']]

        else:
            _sample_corrected_normalized = copy.copy(self.data['sample']['data'])
            _ob_corrected_normalized = copy.copy(self.data['ob']['data'])
            
        self.data['sample']['data'] = _sample_corrected_normalized
        self.data['ob']['data'] = _ob_corrected_normalized
            
        # if the number of sample and ob do not match, use mean of obs
        nbr_sample = len(self.data['sample']['file_name'])
        nbr_ob = len(self.data['ob']['file_name'])
        if (nbr_sample != nbr_ob) or force_mean_ob: # work with mean ob
            _ob_corrected_normalized = np.mean(_ob_corrected_normalized, axis=0)
            self.data['ob']['data_mean'] = _ob_corrected_normalized
            _working_ob = _ob_corrected_normalized.copy()
            _working_ob[_working_ob == 0] = np.NaN

            if notebook:
                # turn on progress bar
                _message = "Normalization"
                box1 = widgets.HBox([widgets.Label(_message,
                                                   layout=widgets.Layout(width='20%')),
                                     widgets.IntProgress(max=len(self.data['sample']['data']))])
                display(box1)
                w1 = box1.children[1]    

            normalized_data = []
            for _index, _sample in enumerate(self.data['sample']['data']):
                _norm = np.divide(_sample, _working_ob)
                _norm[np.isnan(_norm)] = 0
                _norm[np.isinf(_norm)] = 0
                normalized_data.append(_norm)

                if notebook:
                    w1.value = _index+1                
            
        else: # 1 ob for each sample
            # produce normalized data
            sample_ob = zip(self.data['sample']['data'], self.data['ob']['data'])

            if notebook:
                # turn on progress bar
                _message = "Normalization"
                box1 = widgets.HBox([widgets.Label(_message,
                                                   layout=widgets.Layout(width='20%')),
                                     widgets.IntProgress(max=len(self.data['sample']['data']))])
                display(box1)
                w1 = box1.children[1]    

            normalized_data = []
            for _index, [_sample, _ob] in enumerate(sample_ob):
                _working_ob = _ob.copy()
                _working_ob[_working_ob == 0] = np.NaN
                _norm = np.divide(_sample, _working_ob)
                _norm[np.isnan(_norm)] = 0
                _norm[np.isinf(_norm)] = 0
                normalized_data.append(_norm)
                
                if notebook:
                    w1.value = _index+1                                

        self.data['normalized'] = normalized_data

        return True
    
    def data_loaded_have_matching_shape(self):
        '''check that data loaded have the same shape
        
        Returns:
        =======
        bool: result of the check
        '''
        _shape_sample = self.data['sample']['shape']
        _shape_ob = self.data['ob']['shape']
        
        if not (_shape_sample == _shape_ob):
            return False
        
        _shape_df = self.data['df']['shape']
        if not np.isnan(_shape_df['height']):
            if not (_shape_sample == _shape_df):
                return False
            
        return True
    
    def __roi_fit_into_sample(self, roi=[]):
        '''check if roi is within the dimension of the image
        
        Returns:
        ========
        bool: True if roi is within the image dimension
        
        '''
        [sample_height, sample_width] = np.shape(self.data['sample']['data'][0])
        
        [_x0, _y0, _x1, _y1] = [roi.x0, roi.y0, roi.x1, roi.y1]
        if (_x0 < 0) or (_x1 >= sample_width):
            return False
        
        if (_y0 < 0) or (_y1 >= sample_height):
            return False

        return True
    
    def df_correction(self, force=False):
        '''dark field correction of sample and ob
        
        Parameters
        ==========
        force: boolean (default False) that if True will force the df correction to occur, even if it had been
        run before with the same data set

        sample_df_corrected = sample - DF
        ob_df_corrected = OB - DF

        '''
        if not force:
            if self.__exec_process_status['df_correction']:
                return
        self.__exec_process_status['df_correction'] = True
        
        if not self.data['sample']['data'] == []:
            self.__df_correction(data_type='sample')
            
        if not self.data['ob']['data'] == []:
            self.__df_correction(data_type='ob')
    
    def __df_correction(self, data_type='sample'):
        '''dark field correction
        
        Parameters:
           data_type: string ['sample','ob]
        '''
        if not data_type in ['sample', 'ob']:
            raise KeyError("Wrong data type passed. Must be either 'sample' or 'ob'!")

        if self.data['df']['data'] == []:
            return
        
        if self.data['df']['data_average'] == []:
            _df = self.data['df']['data']
            if len(_df) > 1:
                _df = average_df(df=_df)
            self.data['df']['data_average'] = np.squeeze(_df)

        else:
            _df = np.squeeze(self.data['df']['data_average'])

        if np.shape(self.data[data_type]['data'][0]) != np.shape(self.data['df']['data'][0]):
            raise IOError("{} and df data must have the same shape!".format(data_type))
    
        _data_df_corrected = [_data - _df for _data in self.data[data_type]['data']]
        _data_df_corrected = [np.squeeze(_data) for _data in _data_df_corrected]
        self.data[data_type]['data'] = _data_df_corrected
    
    def crop(self, roi=None, force=False):
        ''' Cropping the sample and ob normalized data
        
        Parameters:
        ===========
        roi: ROI object that defines the region to crop
        force: Boolean (default False) that force or not the algorithm to be run more than once
        with the same data set

        Raises:
        =======
        ValueError if sample and ob data have not been normalized yet
        '''
        if (self.data['sample']['data'] == []) or \
           (self.data['ob']['data'] == []):
            raise IOError("We need sample and ob Data !")

        if not type(roi) == ROI:
            raise ValueError("roi must be of type ROI")

        if not force:
            if self.__exec_process_status['crop']:
                return
        self.__exec_process_status['crop'] = True
        
        _x0 = roi.x0
        _y0 = roi.y0
        _x1 = roi.x1
        _y1 = roi.y1
        
        new_sample = [_data[_y0:_y1+1, _x0:_x1+1] for 
                      _data in self.data['sample']['data']]
        self.data['sample']['data'] = new_sample        
       
        new_ob = [_data[_y0:_y1+1, _x0:_x1+1] for 
                  _data in self.data['ob']['data']]
        self.data['ob']['data'] = new_ob        
        
        if not (self.data['df']['data'] == []):
            new_df = [_data[_y0:_y1+1, _x0:_x1+1] for 
                      _data in self.data['df']['data']]
            self.data['df']['data'] = new_df
            
        if not (self.data['normalized'] == []):
            new_normalized = [_data[_y0:_y1+1, _x0:_x1+1] for 
                              _data in self.data['normalized']]
            self.data['normalized'] = new_normalized        
        
        return True
    
    def export(self, folder='./', data_type='normalized', file_type='tif'):
        '''export all the data from the type specified into a folder
        
        Parameters:
        ===========
        folder: String (default is './') where to create all the images. Folder must exist otherwise an error is raised
        data_type: String (default is 'normalized'). Must be one of the following 'sample','ob','df','normalized'
        file_type: String (default is 'tif') format in which to export the data. Must be either 'tif' or 'fits'

        Raises:
        =======
        IOError if the folder does not exist
        KeyError if data_type can not be found in the list ['normalized','sample','ob','df']

        '''
        if not os.path.exists(folder):
            raise IOError("Folder '{}' does not exist!".format(folder))

        if not data_type in ['normalized','sample','ob','df']:
            raise KeyError("data_type '{}' is wrong".format(data_type))

        data = []
        prefix = ''
        if data_type == 'normalized':
            data = self.get_normalized_data()
            prefix = 'normalized'
            data_type = 'sample'
        else:
            data = self.data[data_type]['data']

        if data ==[]:
            return False
        metadata = self.data[data_type]['metadata']

        list_file_name_raw = self.data[data_type]['file_name']
        self.__create_list_file_names(initial_list=list_file_name_raw,
                                      output_folder = folder,
                                      prefix=prefix,
                                      suffix=file_type)
        
        self.__export_data(data=data,
                           metadata=metadata,
                           output_file_names = self._export_file_name,
                           suffix=file_type)
        
    
    def __export_data(self, data=[], metadata=[], output_file_names=[], suffix='tif'):
        '''save the list of files with the data specified
        
        Parameters:
        ===========
        data: numpy array that contains the array of data to save
        output_file_names: numpy array of string of full file names        
        suffix: String (default is 'tif') format in which the file will be created
        '''
        name_data_metadata_array = zip(output_file_names, data, metadata)
        for _file_name, _data, _metadata in name_data_metadata_array:
            if suffix == 'tif':
                make_tif(data=_data, metadata=_metadata, file_name=_file_name)
            elif suffix == 'fits':
                make_fits(data=_data, file_name=_file_name)
    
    def __create_list_file_names(self, initial_list=[], output_folder='', prefix='', suffix=''):
        '''create a list of the new file name used to export the images
        
        Parameters:
        ==========
        initial_list: array of full file name 
           ex: ['/users/me/image001.tif',/users/me/image002.tif',/users/me/image003.tif']
        output_folder: String (default is ./ as specified by calling function) where we want to create the data
        prefix: String. what to add to the output file name in front of base name
           ex: 'normalized' will produce 'normalized_image001.tif'
        suffix: String. extenstion to file. 'tif' for TIFF and 'fits' for FITS
        '''
        _base_name = [os.path.basename(_file) for _file in initial_list]
        _raw_name = [os.path.splitext(_file)[0] for _file in _base_name]
        _prefix = ''
        if prefix:
            _prefix = prefix + '_'
        full_file_names = [os.path.join(output_folder, _prefix + _file + '.' + suffix) for _file in _raw_name]
        self._export_file_name = full_file_names
    
    def get_normalized_data(self):
        '''return the normalized data'''
        return self.data['normalized']

    def get_sample_data(self):
        '''return the sample data'''
        return self.data['sample']['data']
   
    def get_ob_data(self):
        '''return the ob data'''
        return self.data['ob']['data']
    
    def get_df_data(self):
        '''return the df data'''
        return self.data['df']['data']    


#!/usr/bin/env python
from setuptools import setup, find_packages
import NeuNorm

setup(
    name = "NeuNorm",
    version = NeuNorm.__version__,
    author = "Jean Bilheux",
    author_email = "bilheuxjm@ornl.gov",
    packages = find_packages(exclude=['tests', 'notebooks']),
    include_package_data = True,
    test_suite = 'tests',
    install_requires = [
        'numpy',
        'pillow',
        'pathlib',
        'astropy',
        'scipy',
    ],
    dependency_links = [
    ],
    description = "neutron normalization data",
    license = 'BSD',
    keywords = "neutron normalization imaging",
    url = "https://github.com/ornlneutronimaging/NeuNorm",
    classifiers = ['Development Status :: 3 - Alpha',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Intended Audience :: Developers',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.5'],
)


# End of file



import numpy as np


class ROI(object):
    '''class that list the type of ROI available'''

    x0 = np.NaN
    y0 = np.NaN
    x1 = np.NaN
    y1 = np.NaN

    def __init__(self, x0=np.NaN, y0=np.NaN, x1=np.NaN, y1=np.NaN, width=np.NaN, height=np.NaN):
        
        if np.isnan(x0) or np.isnan(y0):
            raise ValueError("x0 and y0 must be provided!")
        
        self.x0 = x0
        self.y0 = y0
        
        if not np.isnan(y1):
            self.y1 = np.max([y0,y1])
            self.y0 = np.min([y0,y1])
        elif not np.isnan(height):
            self.y1 = y0 + height
        else:
            raise ValueError("You must defined either y1 or height!")
            
        if not np.isnan(x1):
            self.x1 = np.max([x0,x1])
            self.x0 = np.min([x0,x1])
        elif not np.isnan(width):
            self.x1 = x0 + width
        else:
            raise ValueError("you must defined either x1 or width!")
        
    
